import 'package:flutter/material.dart';
import 'Tile.dart';
import 'dart:math';

// ignore: must_be_immutable
class TileWidget extends StatelessWidget {
  Tile tile;
  bool isUsable=false;
  TextStyle textStyle = TextStyle(fontSize: 10, fontWeight: FontWeight.bold,color: Colors.red,);
  String nbrText="";
  int index;

  TileWidget(this.tile,this.isUsable,this.nbrText,this.index);

  @override
  Widget build(BuildContext context) {
    return Stack(
        children: <Widget>[
          Container(
              child: Align(
                child: tile.croppedImageTile(),
                alignment: Alignment.center,
              )
          ),
          Center(child: Text(nbrText,style: textStyle,)),
        ]
    );
  }
}

//Decoration des cases cliquables
BoxDecoration myBoxDecoration(bool isUsable) {
  BoxDecoration res=BoxDecoration();
  if(isUsable){
    res=BoxDecoration(
      border: Border.all(
        color: Colors.red,
        width: 5.0,
      ),
    );
  }
  return res;
}

List<Tile> createListTile(double size, Image image){
  List<Tile> res=<Tile>[];

  for(double i=-1.0;i<=1.0;){
    for(double j=-1.0;j<=1.0;){
      res.add(new Tile(image: image, alignment: Alignment(j, i),size: size));
      j+=(2/(size-1));
    }
    i+=(2/(size-1));
  }
  return res;
}

String getRightText(bool play, int i){
  String res="";
  if(play)res= i.toString();
  return res;
}

List<TileWidget>createTileWidgets(double slide, int emptyIndex, List<Tile> listTiles,bool play, bool nombres){
  List<TileWidget> tiles = <TileWidget>[];
  int max=slide.toInt()*slide.toInt();
  int ligne=slide.toInt();
  for(int i=0;i<max;i++){
    if(i==emptyIndex && play){
      tiles.add(TileWidget(Tile(
          image: Image.network("https://upload.wikimedia.org/wikipedia/commons/thumb/6/6e/Fond_blanc.svg/1200px-Fond_blanc.svg.png"),
          alignment: Alignment(-1,-1),size: slide),false,"",i));
    }else if(((i==emptyIndex+1 && i%ligne!=0) ||
        i==emptyIndex+ligne ||
        (i==emptyIndex-1 && (i+1)%ligne!=0)||
        i==emptyIndex-ligne) && play){
      tiles.add(TileWidget(listTiles[i],true,getRightText(nombres, i),i));
    }
    else tiles.add(TileWidget(listTiles[i],false,getRightText(nombres, i),i));
  }
  return tiles;
}

void main() => runApp(new MaterialApp(home: JeuDuTaquin()));

class JeuDuTaquin extends StatefulWidget {
  @override
  State<StatefulWidget> createState() => JeuDuTaquinState();
}

class JeuDuTaquinState extends State<JeuDuTaquin> {
  final TextStyle textStyle = TextStyle(fontSize: 16, fontWeight: FontWeight.bold,);
  final TextStyle victoryStyle = TextStyle(fontSize: 26, fontWeight: FontWeight.bold,color: Colors.red);
  final appBar= AppBar(title: Text('Jeu Du Taquin'),);
  String imageSelect ="Image Aléatoire";
  Image image = Image.network('https://picsum.photos/512/512');
  List<TileWidget> tiles=createTileWidgets(3, 4,createListTile(3,Image.network('https://picsum.photos/512/512')),false,false);
  double _currentSlideValue=3;
  double _currentDifficultyValue=100;
  bool isPlaying=false;
  bool _numbers=false;
  bool _isWon=false;
  int emptyIndex=4;
  int previousEmptyIndex=-1;
  int count=0;
  Random random = new Random();

  @override
  Widget build(BuildContext context) {
    return Scaffold(
        appBar: appBar,
        body: ListView(children: [
          SizedBox(
              width: MediaQuery.of(context).size.width,
              height: MediaQuery.of(context).size.height*0.53,
              child: Container(
                child:GridView.builder(
                  padding: const EdgeInsets.all(10),
                  itemCount: _currentSlideValue.toInt()*_currentSlideValue.toInt(),
                  gridDelegate:SliverGridDelegateWithFixedCrossAxisCount(
                    crossAxisCount: _currentSlideValue.toInt(),
                    crossAxisSpacing: 5.0,
                    mainAxisSpacing: 5.0,
                  ),
                  itemBuilder: (BuildContext context, int index) {
                    return Container(
                      decoration: myBoxDecoration(tiles[index].isUsable),
                      child:InkWell(
                        child: tiles[index],
                        onTap: (){
                          previousEmptyIndex=emptyIndex; //on garde le coup précédent
                          count++; //on incremente le compteur de coup
                          swapTiles(index,emptyIndex);
                          if(isWon()){
                            reset();
                            _isWon=true;
                          }
                        },
                      ),
                    );
                  },
                ),
              ),
          ),
          AnimatedOpacity(
              opacity: _isWon ? 1.0 : 0.0,
              duration: Duration(milliseconds: 2000),
              child: Container(height: 40,child:Align(
                child:Text("VICTOIRE !",style: victoryStyle,),
                alignment: Alignment.center,
              ))
          ),
          !isPlaying? Row(children: <Widget>[
            Container(width: 90,child:Text("Image:", style: textStyle,textAlign: TextAlign.center,),padding: EdgeInsets.fromLTRB(8, 0, 0, 0),),
            Container(height:((MediaQuery.of(context).size.height)/7)-appBar.preferredSize.height,width: MediaQuery.of(context).size.width-115,child:
              DropdownButton<String>(
                value: imageSelect,
                icon: Icon(Icons.arrow_circle_down_outlined, color: Colors.lightBlueAccent,),
                iconSize: 20,
                elevation: 16,
                style: TextStyle(color: Colors.blueGrey,fontSize: 18),
                underline: Container(
                  height: 2,
                  color: Colors.blue,
                ),
                onChanged: (String newValue) {
                  setState(() {
                    imageSelect = newValue;
                    if(imageSelect=="Image Aléatoire") {
                      image=Image.network('https://picsum.photos/512/512');
                    }else image=Image.asset("assets/"+imageSelect+".jpg");

                    tiles=createTileWidgets(_currentSlideValue,emptyIndex,createListTile(_currentSlideValue,image),isPlaying,_numbers);
                  });
                },
                items: <String>['Image Aléatoire', 'Montgolfières','Plage','Ponton','Route', 'Ville',]
                    .map<DropdownMenuItem<String>>((String value) {
                  return DropdownMenuItem<String>(
                    value: value,
                    child: SizedBox(width: MediaQuery.of(context).size.width-135,child:Text(value,textAlign: TextAlign.center,)),
                  );
                }).toList(),
              )
              ),
            ]
          ): new Container(),
          isPlaying? Container(height: 40,child:Align(
            child:Text("Compteur de coups: "+count.toString(),style: textStyle,),
            alignment: Alignment.center,
          )): new Container(),
          isPlaying? Align(
            child:RaisedButton(
              onPressed: () {
                if(previousEmptyIndex!=-1){
                  count--;
                  swapTiles(previousEmptyIndex, emptyIndex);
                }
              },
              child: const Text('Annuler Action', style: TextStyle(fontSize: 20)),
            ),
            alignment: Alignment.center,
          ): new Container(),
          !isPlaying? Row(children: <Widget>[
            Container(width: 90,child:Text('Mélanges:', style: textStyle,textAlign: TextAlign.center,),padding: EdgeInsets.fromLTRB(8, 0, 0, 0),),
            Container(height:((MediaQuery.of(context).size.height)/7)-appBar.preferredSize.height,width: MediaQuery.of(context).size.width-90,child: Slider(
                value: _currentDifficultyValue,
                min: 20, max: 220, divisions: 10,
                label: _currentDifficultyValue.round().toString(),
                onChanged: (double value) {
                  setState(() {
                    if(!isPlaying)_currentDifficultyValue = value;
                  });
                }),
            ),
          ]
          ): new Container(),
          !isPlaying? Row(children: <Widget>[
            Container(width: 90,child:Text('Taille:', style: textStyle,textAlign: TextAlign.center,),padding: EdgeInsets.fromLTRB(8, 0, 0, 0),),
            Container(height:((MediaQuery.of(context).size.height)/7)-appBar.preferredSize.height,width: MediaQuery.of(context).size.width-90,child: Slider(
                value: _currentSlideValue,
                min: 2, max: 7, divisions: 5,
                label: _currentSlideValue.round().toString(),
                onChanged: (double value) {
                  setState(() {
                    if(!isPlaying){
                      _currentSlideValue = value;
                      if(_currentSlideValue%2==1)emptyIndex=((_currentSlideValue*_currentSlideValue)-1)~/2; //si impair alors milieu
                      else emptyIndex=random.nextInt((_currentSlideValue*_currentSlideValue).toInt()); //sinon aleatoire
                      tiles=createTileWidgets(_currentSlideValue,emptyIndex,createListTile(_currentSlideValue,image),isPlaying,_numbers);
                    }
                  });
                }),
              ),
            ]
          ): new Container(),
          !isPlaying? Row(mainAxisSize: MainAxisSize.min,
              children: <Widget>[
                Container(width: 100,child:Text('Afficher les nombres:', style: textStyle, textAlign: TextAlign.center,),padding: EdgeInsets.fromLTRB(8, 0, 0, 0),),
                Container(height:((MediaQuery.of(context).size.height)/7)-appBar.preferredSize.height,width: 70,child: Switch(
                    value: _numbers,
                    onChanged: (value) {
                      setState(() {
                        if(!isPlaying) {
                          _numbers = !_numbers;
                          tiles = createTileWidgets(_currentSlideValue, emptyIndex, createListTile(_currentSlideValue,image), isPlaying, _numbers);
                          }
                        }
                      );
                    }
                  ),
                ),
              ]
          ): new Container(),
      ],),
      floatingActionButtonLocation: FloatingActionButtonLocation.endFloat,
      floatingActionButton: new FloatingActionButton(
        onPressed:(){
          setState(() {
            if(!isPlaying){ //lancement du jeu
              isPlaying=true;
              _isWon=false;
              tiles=createTileWidgets(_currentSlideValue,emptyIndex,createListTile(_currentSlideValue,image),isPlaying,_numbers);
              shuffle(_currentDifficultyValue.toInt());
              }else reset();
            }
          );
        },
        tooltip: 'Play',
        child: new Icon(isPlaying? Icons.cancel:Icons.play_arrow),
      ),
    );
  }

  swapTiles(int current, int empty) {
    setState(() {
      //on retire les bordures rouges apres un deplacement
      for(int i=0;i<tiles.length;i++){
        if(i!=empty ) tiles[i].isUsable=false;
      }

      //deplacement de la case
      if(current==empty+1 && current%_currentSlideValue!=0){
        emptyIndex=current;
        tiles.insert(empty, tiles.removeAt(current));
      }else if(current==empty-1  && empty%_currentSlideValue!=0){
        emptyIndex=current;
        tiles.insert(empty, tiles.removeAt(current));
      }
      else if(current==empty+_currentSlideValue){
        emptyIndex=current;
        tiles.insert(current, tiles.removeAt(empty));
        tiles.insert(empty, tiles.removeAt(current-1));
      }else if(current==empty-_currentSlideValue){
        emptyIndex=current;
        tiles.insert(current, tiles.removeAt(empty));
        tiles.insert(empty, tiles.removeAt(current+1));
      }else print("Can't move this");

      //ajout des bordures rouges
      if(emptyIndex>=_currentSlideValue) tiles[emptyIndex-_currentSlideValue.toInt()].isUsable=true; //haut
      if(emptyIndex<(_currentSlideValue*(_currentSlideValue-1))) tiles[emptyIndex+_currentSlideValue.toInt()].isUsable=true; //bas
      if((emptyIndex+1)%_currentSlideValue!=0)tiles[emptyIndex+1].isUsable=true; //droite
      if(emptyIndex%_currentSlideValue!=0)tiles[emptyIndex-1].isUsable=true;//gauche
    });
  }

  shuffle(int difficulty){
    Random random = new Random();
    for (int i=0;i<difficulty;i++){
      switch (random.nextInt(4)){
        case 0: {
          if(emptyIndex>=_currentSlideValue) {
              swapTiles(emptyIndex - _currentSlideValue.toInt(), emptyIndex);
            }
          else i--;
        }
        break;

        case 1: {
          if((emptyIndex+1)%_currentSlideValue!=0) {
              swapTiles(emptyIndex + 1, emptyIndex);
            }
          else i--;
        }
        break;

        case 2: {
          if(emptyIndex<(_currentSlideValue*(_currentSlideValue-1))) {
              swapTiles(emptyIndex + _currentSlideValue.toInt(), emptyIndex);
            }
          else i--;
        }
        break;

        case 3: {
          if(emptyIndex%_currentSlideValue!=0) {
              swapTiles(emptyIndex - 1, emptyIndex);
            }
          else i--;
        }
        break;

        default: {}
        break;
      }
    }
  }

  reset(){
    isPlaying=false;
    count=0;
    previousEmptyIndex=-1;
    tiles=createTileWidgets(_currentSlideValue,emptyIndex,createListTile(_currentSlideValue,image),isPlaying,_numbers);
  }

  bool isWon(){
    bool res=true;
    int max=_currentSlideValue.toInt()*_currentSlideValue.toInt();
    for(int i=0;i<max;i++){
      if(tiles[i].index!=i)res=false;
    }
    return res;
  }

}